//-----------------------------------------------------------------------------
// Copyright Notice
//
//   Copyright 2002 Sandia Corporation. Under the terms
//   of Contract DE-AC04-94AL85000 with Sandia Corporation, the U.S.
//   Government retains certain rights in this software.
//
//    Xyce(TM) Parallel Electrical Simulator
//    Copyright (C) 2002-2017 Sandia Corporation
//
//    This program is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program.  If not, see <http://www.gnu.org/licenses/>.
//-----------------------------------------------------------------------------

//-------------------------------------------------------------------------
//
// Purpose        :  Provide a linking device for coupling to Alegra
//                   (and other) simulations.
//
// Special Notes  :  The external simulator is responsible for passing
//                   in the number of internal variables in an initialization
//                   step, and for providing F, Q, B (and their derivatives)
//                   each time Xyce is asked to simulate.  
//
// Creator        : Tom Russo, SNL, Electrical and Microsystems Modeling
//
// Creation Date  : 02/28/2017
//
//
//
//
//-------------------------------------------------------------------------
#include <Xyce_config.h>

// ---------- Standard Includes ----------

#include <sstream>

// ----------   Xyce Includes   ----------
#include <N_DEV_DeviceOptions.h>
#include <N_DEV_DeviceMaster.h>
#include <N_DEV_ExternData.h>
#include <N_DEV_MatrixLoadData.h>
#include <N_DEV_SolverState.h>
#include <N_DEV_GeneralExternal.h>
#include <N_DEV_Message.h>
#include <N_ERH_ErrorMgr.h>

#include <N_LAS_Vector.h>
#include <N_LAS_Matrix.h>
#include <N_UTL_FeatureTest.h>

#include <N_DEV_VectorComputeInterface.h>

namespace Xyce {
namespace Device {
namespace GeneralExternal {

void Traits::loadInstanceParameters(ParametricData<GeneralExternal::Instance> &p)
{
}

void Traits::loadModelParameters(ParametricData<GeneralExternal::Model> &p)
{}


// Class Instance
//-----------------------------------------------------------------------------
// Function      : Instance::processParams
// Purpose       :
// Special Notes :
// Scope         : public
// Creator       : Tom Russo, SNL, Electrical Models and Simulation
// Creation Date : 2/28/2017
//-----------------------------------------------------------------------------
bool Instance::processParams()
{
  // If there are any time dependent parameters, set their values at for
  // the current time.

  // now set the temperature related stuff.
  //updateTemperature(temp);

  return true;
}

//-----------------------------------------------------------------------------
// Function      : Instance::updateTemperature
// Purpose       :
// Special Notes :
// Scope         : public
// Creator       : Tom Russo, SNL, Electrical Models and Simulation
// Creation Date : 2/28/2017
//-----------------------------------------------------------------------------
bool Instance::updateTemperature ( const double & temp)
{
  bool bsuccess = true;
  return bsuccess;
}

//-----------------------------------------------------------------------------
// Function      : Instance::Instance
// Purpose       : constructor
// Special Notes :
// Scope         : public
// Creator       : Tom Russo, SNL, Electrical Models and Simulation
// Creation Date : 2/28/2017
//-----------------------------------------------------------------------------
Instance::Instance(
  const Configuration & configuration,
  const InstanceBlock &         IB,
  Model &                       Miter,
  const FactoryBlock &          factory_block)

  : DeviceInstance(IB, configuration.getInstanceParameters(), factory_block),
    model_(Miter),
    vciPtr_(0)
{
  numExtVars   = IB.numExtVars;  // we have as many as were specified on the
                                 // instance line
  numIntVars=0;
  numStateVars=0;

  // Set params to constant default values:
  setDefaultParams ();

  // Set params according to instance line and constant defaults from metadata:
  setParams (IB.params);

  // Set any non-constant parameter defaults:

  // Calculate any parameters specified as expressions:
  updateDependentParameters();

  // calculate dependent (ie computed) params and check for errors:
  processParams ();
}

//-----------------------------------------------------------------------------
// Function      : Instance::~Instance
// Purpose       : destructor
// Special Notes :
// Scope         : public
// Creator       : Tom Russo, SNL, Electrical Models and Simulation
// Creation Date : 2/28/2017
//-----------------------------------------------------------------------------
Instance::~Instance()
{
}

//-----------------------------------------------------------------------------
// Function      : Instance::setupDenseJacStamp_
// Purpose       : Utility function to set up the jacobian stamp
// Special Notes : Just makes a dense jacobian stamp for now
// Scope         : private
// Creator       : Tom Russo, SNL, Electrical Models and Simulation
// Creation Date : 2/28/2017
//-----------------------------------------------------------------------------
void Instance::setupDenseJacStamp_()
{
 
  int numNodes = numExtVars+numIntVars;

  jacStamp_.resize(numNodes);
  for (int i=0; i<numNodes; ++i)
  {
    jacStamp_[i].resize(numNodes);
    for (int j=0; j<numNodes; ++j)
    {
      jacStamp_[i][j]=j;
    }
  }
}


//-----------------------------------------------------------------------------
// Function      : Instance::registerLIDs
// Purpose       :
// Special Notes :
// Scope         : public
// Creator       : Tom Russo, SNL, Electrical Models and Simulation
// Creation Date : 2/28/2017
//-----------------------------------------------------------------------------
void Instance::registerLIDs(const std::vector<int> & intLIDVecRef,
                            const std::vector<int> & extLIDVecRef)
{
  AssertLIDs(intLIDVecRef.size() == numIntVars);
  AssertLIDs(extLIDVecRef.size() == numExtVars);

  if (DEBUG_DEVICE && isActive(Diag::DEVICE_PARAMETERS))
  {
    Xyce::dout() << std::endl << section_divider << std::endl;
    Xyce::dout() << "  Instance::registerLIDs" << std::endl;
    Xyce::dout() << "  name = " << getName() << std::endl;
  }

  int numNodes = numExtVars + numIntVars;

  // copy over the global ID lists.
  intLIDVec = intLIDVecRef;
  extLIDVec = extLIDVecRef;

  li_Nodes_.resize(numNodes);

  int extVar=0;
  int intVar=0;
  int theVar=0;

  // First comes the external nodes
  for (theVar=0; theVar<numExtVars; ++theVar)
  {
    li_Nodes_[theVar] = extLIDVec[theVar];
  }
  // now the internals

  for (theVar=0; theVar<+numIntVars; ++theVar)
  {
      li_Nodes_[theVar+numExtVars] = intLIDVec[theVar];
  }


  if (DEBUG_DEVICE && isActive(Diag::DEVICE_PARAMETERS) )
  {
    for (int i=0; i<numIntVars+numExtVars;++i)
    {
      Xyce::dout() << "LID var " << i << " ";
      if (i>=numExtVars)
        Xyce::dout() << " (internal) ";
      Xyce::dout() << ": "<< li_Nodes_[i] << std::endl;
    }
    Xyce::dout() << section_divider << std::endl;
  }
}

//-----------------------------------------------------------------------------
// Function      : Instance::loadNodeSymbols
// Purpose       :
// Special Notes :
// Scope         : public
// Creator       : Tom Russo, SNL, Electrical Models and Simulation
// Creation Date : 2/28/2017
//-----------------------------------------------------------------------------
void Instance::loadNodeSymbols(Util::SymbolTable &symbol_table) const
{
  if (numIntVars != 0)
  {
    for (int i = 0; i < numIntVars; ++i)
    {
      std::ostringstream ost;
      ost << "InternalNode_" << i;
      addInternalNode(symbol_table, li_Nodes_[numExtVars+i],
                      getName(), ost.str());
    }
  }
}

//-----------------------------------------------------------------------------
// Function      : Instance::registerStateLIDs
// Purpose       :
// Special Notes :
// Scope         : public
// Creator       : Tom Russo, SNL, Electrical Models and Simulation
// Creation Date : 2/28/2017
//-----------------------------------------------------------------------------
void Instance::registerStateLIDs( const std::vector<int> & staLIDVecRef )
{
  AssertLIDs(staLIDVecRef.size() == numStateVars);
}

//-----------------------------------------------------------------------------
// Function      : Instance::jacobianStamp
// Purpose       :
// Special Notes :
// Scope         : public
// Creator       : Tom Russo, SNL, Electrical Models and Simulation
// Creation Date : 2/28/2017
//-----------------------------------------------------------------------------
const std::vector< std::vector<int> > & Instance::jacobianStamp() const
{
  return jacStamp_;
}

//-----------------------------------------------------------------------------
// Function      : Instance::registerJacLIDs
// Purpose       :
// Special Notes :
// Scope         : public
// Creator       : Tom Russo, SNL, Electrical Models and Simulation
// Creation Date : 2/28/2017
//-----------------------------------------------------------------------------
void Instance::registerJacLIDs( const std::vector< std::vector<int> > & jacLIDVec )
{
  if (jacStamp_.empty())
  {
    DevelFatal(*this) << "Improper setup.  No Jacobian stamp has been set."
                      << std::endl
                      << "The General External device can only be used through the "
                      << "GenCouplingSimulator interface, and cannot be used in normal Xyce runs. "
                      << std::endl
                      << "If you are calling this device through the GenCouplingInterface, be sure "
                      << "that you are setting the number of interal variables with setNumInternalVariables"
                      << "and associating  a vector loader with setVectorLoader." << std::endl;
  }    

  DeviceInstance::registerJacLIDs( jacLIDVec );

  int numVars=numExtVars+numIntVars;
   A_Equ_NodeOffsets_.resize(numVars);
  for (int equ=0; equ < numVars; ++equ)
  {
    A_Equ_NodeOffsets_[equ].resize(jacStamp_[equ].size());
    for (int node=0; node < jacStamp_[equ].size(); ++node)
    {
      A_Equ_NodeOffsets_[equ][node] = jacLIDVec[equ][node];
    }
  }

}

//-----------------------------------------------------------------------------
// Function      : Instance::updatePrimaryState
// Purpose       :
// Special Notes :
// Scope         : public
// Creator       : Tom Russo, SNL, Electrical Models and Simulation
// Creation Date : 2/28/2017
//-----------------------------------------------------------------------------
bool Instance::updatePrimaryState()
{
  bool bsuccess = true;

  bsuccess = updateIntermediateVars();

  return bsuccess;
}

//-----------------------------------------------------------------------------
// Function      : Instance::updateSecondaryState
// Purpose       :
// Special Notes :
// Scope         : public
// Creator       : Tom Russo, SNL, Electrical Models and Simulation
// Creation Date : 2/28/2017
//-----------------------------------------------------------------------------
bool Instance::updateSecondaryState ()
{
  bool bsuccess = true;

  return bsuccess;
}


//-----------------------------------------------------------------------------
// Function      : Instance::updateIntermediateVars
// Purpose       :
// Special Notes :
// Scope         : public
// Creator       : Tom Russo, SNL, Electrical Models and Simulation
// Creation Date : 2/28/2017
//-----------------------------------------------------------------------------
bool Instance::updateIntermediateVars ()
{

  bool bsuccess=true;
  int numVars=numIntVars+numExtVars;
  double * solVector = extData.nextSolVectorRawPtr;
  if (vciPtr_)
  {
    if (solutionVars_.empty())
      solutionVars_.resize(numVars);

    // Copy out the solution variables we need.  The ordering
    // is already correct, because of the way we constructed
    // li_Nodes
    for (int i=0;i<numVars;i++)
      solutionVars_[i] = solVector[li_Nodes_[i]];

    // Call back to the computation object and get this devices
    // contributions.  IT IS THAT FUNCTION'S RESPONSIBILITY TO SIZE THESE
    // VECTORS APPROPRIATELY!
    bsuccess=vciPtr_->computeXyceVectors(solutionVars_,
                                         getSolverState().currTime_,
                                         FVec_, QVec_, BVec_,
                                         dFdXMat_, dQdXMat_);

    // NOTE:  This sanity checking could be costly, and should
    // only be enabled with DEBUG_DEVICE turned on.
    if (DEBUG_DEVICE)
    {
      if (bsuccess)
      {
        // Sanity check sizes.  Each vector MUST be either 0 length (indicating
        // no contribution to that vector) *OR* numVars long.  And at least ONE
        // of the vectors MUST be non-zero length.
        if ((FVec_.empty() && QVec_.empty() && BVec_.empty())
            || (FVec_.size()!=numVars && QVec_.size()!=numVars
                && BVec_.size()!= numVars)
            )
        {
          UserError(*this) << "At least one of the F, Q, or B matrices must have length " << numVars;
          bsuccess=false;
        }
        else
        {
          if ( !((FVec_.size() == numVars || FVec_.size() == 0) &&
                 (QVec_.size() == numVars || QVec_.size() == 0) &&
                 (BVec_.size() == numVars || BVec_.size() == 0)))
          {
            UserError(*this) << "F, Q, or B matrices must have length "
                             << numVars << " or zero. "
                             << " F length=" << FVec_.size()
                             << " Q length=" << QVec_.size()
                             << " B length=" << BVec_.size();
            bsuccess=false;
          }
          else
          {
            if ( dFdXMat_.size() != FVec_.size() || dQdXMat_.size() != QVec_.size())
            {
              UserError(*this) << "dFdX matrix has " << dFdXMat_.size()
                               << " rows "
                               << " and should match F vector length "
                               << FVec_.size()
                               << ", dQdX matrix has " << dQdXMat_.size()
                               << " rows "
                               << " and should match Q vector length "
                               << QVec_.size();
              bsuccess=false;
            }
            else
            {
              // Here's the potentially costly one
              // All the vectors are the right sizes, all the matrices
              // have the right rows --- do all the rows have the right
              // lengths?
              for (int i=0; i<dFdXMat_.size();i++)
              {
                if (dFdXMat_[i].size() != FVec_.size())
                {
                  UserError(*this) << " F matrix row " << i << " has "
                                   << dFdXMat_[i].size()
                                   << " must have " << FVec_.size();
                  bsuccess=false;
                }
              }
              for (int i=0; i<dQdXMat_.size();i++)
              {
                if (dQdXMat_[i].size() != QVec_.size())
                {
                  UserError(*this) << " Q matrix row " << i << " has "
                                   << dQdXMat_[i].size()
                                   << " must have " << QVec_.size();
                  bsuccess=false;
                }
              }
            }
          }
        }
      }      
    }
  }
  else
  {
    bsuccess=false;
    DevelFatal(*this) << "Improper setup.  No vector loader for this has been assigned."
                      << std::endl
                      << "The General External device can only be used through the"
                      << "GenCouplingSimulator interface, and cannot be used in normal Xyce runs."
                      << std::endl
                      << "If you are calling this device through the GenCouplingInterface, be sure"
                      << "that you are setting the number of interal variables with setNumInternalVariables "
                      << "and associating  a vector loader with setVectorLoader." << std::endl;
      
  }

  return bsuccess;
}

//-----------------------------------------------------------------------------
// Function      : Instance::loadDAEQVector
//
// Purpose       : Loads the Q-vector contributions for a single
//                 instance.
//
// Special Notes : The "Q" vector is part of a standard DAE formalism in
//                 which the system of equations is represented as:
//
//                 f(x) = dQ(x)/dt + F(x) - B(t) = 0
//
// Scope         : public
// Creator       : Tom Russo, SNL, Electrical Models and Simulation
// Creation Date : 2/28/2017
//-----------------------------------------------------------------------------
bool Instance::loadDAEQVector ()
{
  bool bsuccess = true;
  int numVars = numExtVars+numIntVars;

  Linear::Vector * daeQVecPtr = extData.daeQVectorPtr;

  if (QVec_.size() != 0)
  {
    for (int i=0; i<numVars; i++)
      (*daeQVecPtr)[li_Nodes_[i]] += QVec_[i];
  }
  
  return bsuccess;
}



//-----------------------------------------------------------------------------
// Function      : Instance::loadDAEFVector
//
// Purpose       : Loads the F-vector contributions for a single
//                 instance.
//
// Special Notes :
//
// Scope         : public
// Creator       : Tom Russo, SNL, Electrical Models and Simulation
// Creation Date : 2/28/2017
//-----------------------------------------------------------------------------
bool Instance::loadDAEFVector ()
{
  bool bsuccess=true;

  Linear::Vector * daeFVecPtr = extData.daeFVectorPtr;
  int numVars = numExtVars+numIntVars;

  if (FVec_.size() != 0)
  {
    for (int i=0; i<numVars; i++)
      (*daeFVecPtr)[li_Nodes_[i]] += FVec_[i];
  }

  return bsuccess;
}

//-----------------------------------------------------------------------------
// Function      : Instance::loadDAEBVector
//
// Purpose       : Loads the B-vector contributions for a single
//                 instance.
//
// Special Notes :
//
// Scope         : public
// Creator       : Tom Russo, SNL, Electrical Models and Simulation
// Creation Date : 2/28/2017
//-----------------------------------------------------------------------------
bool Instance::loadDAEBVector ()
{
  bool bsuccess=true;

  Linear::Vector * daeBVecPtr = extData.daeBVectorPtr;
  int numVars = numExtVars+numIntVars;

  if (BVec_.size() != 0)
  {
    for (int i=0; i<numVars; i++)
      (*daeBVecPtr)[li_Nodes_[i]] += BVec_[i];
  }

  return bsuccess;
}

//-----------------------------------------------------------------------------
// Function      : Instance::loadDAEdQdx
//
// Purpose       : Loads the Q-vector contributions for a single
//                 instance.
//
// Special Notes:  We are assuming a dense jacstamp here!
//
// Scope         : public
// Creator       : Tom Russo, SNL, Electrical Models and Simulation
// Creation Date : 2/28/2017
//-----------------------------------------------------------------------------
bool Instance::loadDAEdQdx ()
{
  bool bsuccess = true;
  Linear::Matrix * dQdxMatPtr = extData.dQdxMatrixPtr;
  int numVars = numExtVars+numIntVars;

  if (QVec_.size() != 0)
  {
    for (int row=0; row<numVars; row++)
    {
      for (int col=0; col<A_Equ_NodeOffsets_[row].size(); col++)
      {
        (*dQdxMatPtr)[li_Nodes_[row]][A_Equ_NodeOffsets_[row][col]]
          += dQdXMat_[row][jacStamp_[row][col]];
      }
    }
  }
        
  return bsuccess;
}



//-----------------------------------------------------------------------------
// Function      : Instance::loadDAEdFdx ()
//
// Purpose       : Loads the F-vector contributions for a single
//                 instance.
//
// Special Notes:  We are assuming a dense jacstamp here!
//
// Scope         : public
// Creator       : Tom Russo, SNL, Electrical Models and Simulation
// Creation Date : 2/28/2017
//-----------------------------------------------------------------------------
bool Instance::loadDAEdFdx ()
{
  bool bsuccess = true;

  Linear::Matrix * dFdxMatPtr = extData.dFdxMatrixPtr;
  int numVars = numExtVars+numIntVars;

  if (FVec_.size() != 0)
  {
    for (int row=0; row<numVars; row++)
    {
      for (int col=0; col<A_Equ_NodeOffsets_[row].size(); col++)
      {
        (*dFdxMatPtr)[li_Nodes_[row]][A_Equ_NodeOffsets_[row][col]]
          += dFdXMat_[row][jacStamp_[row][col]];
      }
    }
  }

  return bsuccess;
}

//-----------------------------------------------------------------------------
// Function      : Instance::getSolution
//
// Purpose       : Copies this device's solution variables into provide
//                 vector
//
// Special Notes :
//
// Scope         : public
// Creator       : Tom Russo, SNL, Electrical Models and Simulation
// Creation Date : 2/28/2017
//-----------------------------------------------------------------------------
void Instance::getSolution(std::vector<double> &sV)
{
  double * solVector = extData.nextSolVectorRawPtr;
  int numVars = numExtVars+numIntVars;

  if (solutionVars_.empty())
    solutionVars_.resize(numVars);
  
  // Copy out the solution variables we need.  The ordering
  // is already correct, because of the way we constructed
  // li_Nodes
  for (int i=0;i<numVars;i++)
    solutionVars_[i] = solVector[li_Nodes_[i]];

  sV = solutionVars_;
}

//-----------------------------------------------------------------------------
// Function      : Model::processParams
// Purpose       :
// Special Notes :
// Scope         : public
// Creator       : Tom Russo, SNL, Electrical Models and Simulation
// Creation Date : 2/28/2017
//-----------------------------------------------------------------------------
bool Model::processParams ()
{
  return true;
}

//----------------------------------------------------------------------------
// Function      : Model::processInstanceParams
// Purpose       :
// Special Notes :
// Scope         : public
// Creator       : Tom Russo, SNL, Electrical Models and Simulation
// Creation Date : 2/28/2017
//----------------------------------------------------------------------------
bool Model::processInstanceParams()
{

  std::vector<Instance*>::iterator iter;
  std::vector<Instance*>::iterator first = instanceContainer.begin();
  std::vector<Instance*>::iterator last  = instanceContainer.end();

  for (iter=first; iter!=last; ++iter)
  {
    (*iter)->processParams();
  }

  return true;
}

//-----------------------------------------------------------------------------
// Function      : Model::Model
// Purpose       : block constructor
// Special Notes :
// Scope         : public
// Creator       : Tom Russo, SNL, Electrical Models and Simulation
// Creation Date : 2/28/2017
//-----------------------------------------------------------------------------
Model::Model(
  const Configuration & configuration,
  const ModelBlock &    MB,
  const FactoryBlock &  factory_block)
  : DeviceModel(MB, configuration.getModelParameters(), factory_block)
{

  // Set params to constant default values:
  setDefaultParams ();

  // Set params according to .model line and constant defaults from metadata:
  setModParams (MB.params);

  // Set any non-constant parameter defaults:

  // Calculate any parameters specified as expressions:
  updateDependentParameters();

  // calculate dependent (ie computed) params and check for errors:

  processParams ();
}

//-----------------------------------------------------------------------------
// Function      : Model::~Model
// Purpose       : destructor
// Special Notes :
// Scope         : public
// Creator       : Tom Russo, SNL, Electrical Models and Simulation
// Creation Date : 2/28/2017
//-----------------------------------------------------------------------------
Model::~Model ()
{
  std::vector<Instance*>::iterator iter;
  std::vector<Instance*>::iterator first = instanceContainer.begin();
  std::vector<Instance*>::iterator last  = instanceContainer.end();

  for (iter=first; iter!=last; ++iter)
  {
    delete (*iter);
  }

}

// additional Declarations

//-----------------------------------------------------------------------------
// Function      : Model::printOutInstances
// Purpose       : debugging tool.
// Special Notes :
// Scope         : public
// Creator       : Tom Russo, SNL, Electrical Models and Simulation
// Creation Date : 2/28/2017
//-----------------------------------------------------------------------------
std::ostream &Model::printOutInstances(std::ostream &os) const
{
  std::vector<Instance*>::const_iterator iter;
  std::vector<Instance*>::const_iterator first = instanceContainer.begin();
  std::vector<Instance*>::const_iterator last  = instanceContainer.end();

  int i, isize;
  isize = instanceContainer.size();

  os << std::endl;
  os << "Number of Xygra instances: " << isize << std::endl;
  os << "    name=\t\tmodelName\tParameters" << std::endl;
  for (i=0, iter=first; iter!=last; ++iter, ++i)
  {
    os << "  " << i << ": " << (*iter)->getName() << "\t";
    os << getName();
    os << std::endl;
  }

  os << std::endl;

  return os;
}

//-----------------------------------------------------------------------------
// Function      : Model::forEachInstance
// Purpose       : 
// Special Notes :
// Scope         : public
// Creator       : Tom Russo, SNL, Electrical Models and Simulation
// Creation Date : 2/28/2017
//-----------------------------------------------------------------------------
/// Apply a device instance "op" to all instances associated with this
/// model
/// 
/// @param[in] op Operator to apply to all instances.
/// 
/// 
void Model::forEachInstance(DeviceInstanceOp &op) const /* override */ 
{
  for (std::vector<Instance *>::const_iterator it = instanceContainer.begin(); it != instanceContainer.end(); ++it)
    op(*it);
}


Device *Traits::factory(const Configuration &configuration, const FactoryBlock &factory_block)
{

  return new DeviceMaster<Traits>(configuration, factory_block, factory_block.solverState_, factory_block.deviceOptions_);
}

void registerDevice()
{
  Config<Traits>::addConfiguration()
    .registerDevice("genext", 1)
    .registerModelType("genext", 1);
}

} // namespace GeneralExternal
} // namespace Device
} // namespace Xyce
