//-*-c++-*-
//-----------------------------------------------------------------------------
// Copyright Notice
//
//   Copyright 2002 Sandia Corporation. Under the terms
//   of Contract DE-AC04-94AL85000 with Sandia Corporation, the U.S.
//   Government retains certain rights in this software.
//
//    Xyce(TM) Parallel Electrical Simulator
//    Copyright (C) 2002-2017 Sandia Corporation
//
//    This program is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program.  If not, see <http://www.gnu.org/licenses/>.
//-----------------------------------------------------------------------------

//-----------------------------------------------------------------------------
//
// Purpose        : General External Device classes.
//
// Special Notes  :
//
// Creator        : Tom Russo, SNL, Electrical Models and Simulation
// Creation Date  : 03/01/2017
//
//-----------------------------------------------------------------------------

#ifndef Xyce_N_DEV_GeneralExternal_h
#define Xyce_N_DEV_GeneralExternal_h

// ----------   Xyce Includes   ----------
#include <N_DEV_Configuration.h>
#include <N_DEV_DeviceBlock.h>
#include <N_DEV_DeviceInstance.h>
#include <N_DEV_VectorComputeInterface.h>

namespace Xyce {
namespace Device {
namespace GeneralExternal {
class Model;
class Instance;

struct Traits : public DeviceTraits<Model, Instance>
{
  static const char *name() {return "GeneralExternal";}
  static const char *deviceTypeName() {return "GeneralExternal device level 1";}
  static int numNodes() {return 2;}
  static int numOptionalNodes() {return 1000;}
  static bool isLinearDevice() {return false;}

  static Device *factory(const Configuration &configuration, const FactoryBlock &factory_block);
  static void loadModelParameters(ParametricData<Model> &model_parameters);
  static void loadInstanceParameters(ParametricData<Instance> &instance_parameters);
};

//-----------------------------------------------------------------------------
// Class         : Instance
// Purpose       : This is class refers to a single instance of the
//                 GeneralExternal device.  
// Special Notes :
// Creator       : Tom Russo, SNL, Electrical Models and Simulation
// Creation Date : 2/28/2017
//-----------------------------------------------------------------------------

class Instance : public DeviceInstance
{
  friend class ParametricData<Instance>;
  friend class Model;
  friend struct Traits;

public:
  Instance(
     const Configuration &       configuration,
     const InstanceBlock &            IB,
     Model & Miter,
     const FactoryBlock &factory_block);

  ~Instance();

private:
  Instance(const Instance &);
  Instance &operator=(const Instance &);

public:
  void registerLIDs( const std::vector<int> & intLIDVecRef,
                     const std::vector<int> & extLIDVecRef );
  void registerStateLIDs( const std::vector<int> & staLIDVecRef );

  void loadNodeSymbols(Util::SymbolTable &symbol_table) const; // override

  const std::vector< std::vector<int> > & jacobianStamp() const;
  void registerJacLIDs( const std::vector< std::vector<int> > & jacLIDVec );

  bool processParams ();
  bool updateTemperature(const double & temp_tmp);

  bool updateIntermediateVars ();
  bool updatePrimaryState ();
  bool updateSecondaryState ();

  int getNumVars();

  // load functions, residual:
  bool loadDAEQVector ();
  bool loadDAEFVector ();
  bool loadDAEBVector ();

  void auxDAECalculations ();

  // load functions, Jacobian:
  bool loadDAEdQdx ();
  bool loadDAEdFdx ();

  bool setVectorLoader(Xyce::Device::VectorComputeInterface * vciPtr);
  void setNumInternalVars(int numInt);
  void setJacStamp(std::vector< std::vector<int> > & jS);
  void getSolution(std::vector<double> &sV);

protected:
private:
  void setupDenseJacStamp_();

public:
  // iterator reference to the Xygra model which owns this instance.
  // Getters and setters
  Model &getModel() 
  {
    return model_;
  }

  
private:

  Model &       model_;         //< Owning model

private:
  // parameter variables

  // state variables
  // This device has no state

  // local state indices (offsets)
  // This device has no state

  // local solution indices (offsets)
  // This device uses an array of li_ values instead of individually named
  // variables.
  std::vector<int> li_Nodes_;

  // Matrix equation index variables:

  // Offset variables.  Again, this device uses an array instead of
  // discrete variables.
  // A_Equ_NodeOffests[equation][node] is the offset for node in
  // equation
  std::vector< std::vector<int> > A_Equ_NodeOffsets_;

  std::vector< std::vector<int> > jacStamp_;

  Xyce::Device::VectorComputeInterface * vciPtr_;

  /// Vectors and matrices for storing local representations of F,
  /// Q, and B vectors, used for passing to the computeXyceVectors
  /// function of the *vciPtr_ object.
  std::vector<double> solutionVars_;
  std::vector<double> FVec_;
  std::vector<double> QVec_;
  std::vector<double> BVec_;
  std::vector<std::vector<double> > dFdXMat_;
  std::vector<std::vector<double> > dQdXMat_;

};

//-----------------------------------------------------------------------------
// Class         : Model
// Purpose       :
// Special Notes :
// Creator       : Tom Russo, SNL, Electrical Models and Simulation
// Creation Date : 2/17/2017
//-----------------------------------------------------------------------------
class Model : public DeviceModel
{
  typedef std::vector<Instance *> InstanceVector;

  friend class ParametricData<Model>;
  friend class Instance;
  friend struct Traits;
    
public:
  Model(
     const Configuration &       configuration,
     const ModelBlock &          MB,
     const FactoryBlock &        factory_block);
  ~Model();

private:
  Model();
  Model(const Model &);
  Model &operator=(const Model &);

public:
  virtual void forEachInstance(DeviceInstanceOp &op) const /* override */;
    
  virtual std::ostream &printOutInstances(std::ostream &os) const;

  bool processParams ();
  bool processInstanceParams ();

public:
  void addInstance(Instance *instance) 
  {
    instanceContainer.push_back(instance);
  }

private:
  std::vector<Instance*> instanceContainer;

private:

  // Additional Implementation Declarations
};

//----------------------------------------------------------------------------
// Function      : Instance::setVectorLoader
// Purpose       : Associate a VectorCompute object with this device
// Special Notes :
// Scope         : public
// Creator       : Tom Russo, SNL, Electrical Models and Simulation
// Creation Date : 2/28/2017
//----------------------------------------------------------------------------
/// associate a vector loader object pointer with this instance
inline bool Instance::setVectorLoader(Xyce::Device::VectorComputeInterface * vciPtr)
{
  bool bsuccess=true;
  // Because we did not do this in the constructor as is typical for
  // most devices, make sure we get it done now.
  if (jacStamp_.empty())
      setupDenseJacStamp_();

  if (vciPtr)
    vciPtr_=vciPtr;
  else
    bsuccess=false;

  return bsuccess;
}

//----------------------------------------------------------------------------
// Function      : Instance::setNumInternalVars
// Purpose       : Set the number of internal variables for this device
// Special Notes : For God's sake, make sure to call this before
//                 topology needs to query the number!
// Scope         : public
// Creator       : Tom Russo, SNL, Electrical Models and Simulation
// Creation Date : 3/1/2017
//----------------------------------------------------------------------------
inline void Instance::setNumInternalVars(int numInt)
{
  numIntVars=numInt;
}

//----------------------------------------------------------------------------
// Function      : Instance::setJacStamp
// Purpose       : Set the jacobian stamp
// Special Notes : For God's sake, make sure to call this before
//                 topology needs to query the jacStamp!
// Scope         : public
// Creator       : Tom Russo, SNL, Electrical Models and Simulation
// Creation Date : 3/1/2017
//----------------------------------------------------------------------------
///
/// Set the jacobian stamp
///
/// The Jacobian Stamp is a sparse representation of the pattern of non-zero
/// elements of the jacobian associated with a device.
///
/// The jacobian stamp must have as many rows as the device has variables
/// (i.e. numIntVars+numExtVars).
///
/// Each row of the jacStamp is associated with an equation for the device.
/// The number of elements of a row is the number of non-zero Jacobian
/// entries for that equation.
///
/// The values of the elements in a row are the variable numbers for the
/// non-zero elements of the Jacobian.
///
/// Currently, Xyce uses the same jacStamp for both dFdX and dQdX, so
/// the jacobian stamp must be the union of nonzeros for these two
/// matrices.
///
/// A jacobian stamp row may be empty, indicating that the F and Q vector
/// elements for that row have no dependence on any variable, or the vector
/// elements are always zero (as would be the case for control nodes of a
/// controlled source, for example).
///
/// If this function is never called, a fully dense jacobian stamp
/// (in which every equation may depend on every variable) is constructed
/// at the time that setVectorLoader is called.
inline void Instance::setJacStamp(  std::vector< std::vector<int> > & jS)
{
  jacStamp_ = jS;
}

//----------------------------------------------------------------------------
// Function      : Instance::getNumVars
// Purpose       : Return the number of solution vars in a given instance
// Special Notes :
// Scope         : public
// Creator       : Tom Russo, SNL, Electrical Models and Simulation
// Creation Date : 08/27/2008
//----------------------------------------------------------------------------
inline int Instance::getNumVars()
{
  return numExtVars+numIntVars;
}

void registerDevice();

} // namespace Resistor
} // namespace Device
} // namespace Xyce

#endif
