//-----------------------------------------------------------------------------
// Copyright Notice
//
//   Copyright 2002 Sandia Corporation. Under the terms
//   of Contract DE-AC04-94AL85000 with Sandia Corporation, the U.S.
//   Government retains certain rights in this software.
//
//    Xyce(TM) Parallel Electrical Simulator
//    Copyright (C) 2002-2017 Sandia Corporation
//
//    This program is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//   (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program.  If not, see <http://www.gnu.org/licenses/>.
//-----------------------------------------------------------------------------

//-------------------------------------------------------------------------
//
// Purpose        :
//
// Special Notes  :
//
// Creator        : Dave Baur
//
// Creation Date  :
//
//
//
//
//-------------------------------------------------------------------------

#include <Xyce_config.h>

#include <N_ERH_ErrorMgr.h>
#include <N_IO_OutputterSensitivity.h>
#include <N_IO_OutputterTimePrn.h>
#include <N_IO_OutputterTimeCSV.h>
#include <N_IO_OutputterFrequencyPrn.h>
#include <N_IO_OutputterFrequencyCSV.h>
#include <N_IO_OutputterSensitivityPrn.h>
#include <N_IO_OutputterSensitivityTecplot.h>
#include <N_IO_OutputterSensitivityDakota.h>
#include <N_IO_OutputMgr.h>

namespace Xyce {
namespace IO {
namespace Outputter {

//-----------------------------------------------------------------------------
// Function      : enableSensitivityOutput
// Purpose       :
// Special Notes :
// Scope         :
// Creator       : Dave Baur
// Creation Date : June 23, 2014
//-----------------------------------------------------------------------------
void enableSensitivityOutput(Parallel::Machine comm, OutputMgr &output_manager, Analysis::Mode analysis_mode)
{
  std::pair<OutputParameterMap::const_iterator, bool> result = output_manager.findOutputParameter(OutputType::SENS);
  if (result.second)
  {
    for (std::vector<PrintParameters>::const_iterator it = (*result.first).second.begin(), end = (*result.first).second.end(); it != end; ++it) 
    {
      PrintParameters sensitivity_print_parameters = (*it);

      if (analysis_mode == Analysis::ANP_MODE_TRANSIENT)
        sensitivity_print_parameters.variableList_.push_front(Util::Param("TIME", 0.0));
      if (sensitivity_print_parameters.printIndexColumn_)
        sensitivity_print_parameters.variableList_.push_front(Util::Param("INDEX", 0.0));

      output_manager.fixupPrintParameters(comm, sensitivity_print_parameters);

      Outputter::Interface *outputter;
      if (sensitivity_print_parameters.format_ == Format::STD) 
      {
        sensitivity_print_parameters.defaultExtension_ = ".SENS.prn";
        outputter = new Outputter::SensitivityPrn(comm, output_manager, sensitivity_print_parameters);
      }
      else if (sensitivity_print_parameters.format_ == Format::TECPLOT) 
      {
        sensitivity_print_parameters.defaultExtension_ = ".SENS.dat";
        outputter = new Outputter::SensitivityTecPlot(comm, output_manager, sensitivity_print_parameters);
      }
      else if (sensitivity_print_parameters.format_ == Format::DAKOTA) 
      {
        sensitivity_print_parameters.defaultExtension_ = ".SENS.txt";
        outputter = new Outputter::SensitivityDakota(comm, output_manager, sensitivity_print_parameters);
      }
      else
      {
        Report::UserWarning0()
          << "Sensitivity output cannot be written in requested format, using standard format";

        sensitivity_print_parameters.defaultExtension_ = ".SENS.prn";
        outputter = new Outputter::SensitivityPrn(comm, output_manager, sensitivity_print_parameters);
      }

      output_manager.addOutputter(PrintType::SENS, outputter);
    }
  }
}

//-----------------------------------------------------------------------------
// Function      : enableAdjointSensitivityOutput
// Purpose       :
// Special Notes :
// Scope         :
// Creator       : Eric Keiter
// Creation Date : 4/12/2016
//-----------------------------------------------------------------------------
void enableAdjointSensitivityOutput(Parallel::Machine comm, OutputMgr &output_manager, Analysis::Mode analysis_mode)
{
  std::pair<OutputParameterMap::const_iterator, bool> result = output_manager.findOutputParameter(OutputType::TRANADJOINT);
  if (result.second)
  {
    for (std::vector<PrintParameters>::const_iterator it = (*result.first).second.begin(), end = (*result.first).second.end(); it != end; ++it) 
    {
      PrintParameters transientAdjoint_print_parameters = (*it);

      if (analysis_mode == Analysis::ANP_MODE_TRANSIENT)
      {
        transientAdjoint_print_parameters.variableList_.push_front(Util::Param("TIME", 0.0));
      }

      if (transientAdjoint_print_parameters.printIndexColumn_)
      {
        transientAdjoint_print_parameters.variableList_.push_front(Util::Param("INDEX", 0.0));
      }

      output_manager.fixupPrintParameters(comm, transientAdjoint_print_parameters);

      Outputter::Interface *outputter;
      if (transientAdjoint_print_parameters.format_ == Format::STD) 
      {
        transientAdjoint_print_parameters.defaultExtension_ = ".TRADJ.prn";
        outputter = new Outputter::SensitivityPrn(comm, output_manager, transientAdjoint_print_parameters);
      }
      else if (transientAdjoint_print_parameters.format_ == Format::TECPLOT) 
      {
        transientAdjoint_print_parameters.defaultExtension_ = ".TRADJ.dat";
        outputter = new Outputter::SensitivityTecPlot(comm, output_manager, transientAdjoint_print_parameters);
      }
      else if (transientAdjoint_print_parameters.format_ == Format::DAKOTA) 
      {
        transientAdjoint_print_parameters.defaultExtension_ = ".TRADJ.txt";
        outputter = new Outputter::SensitivityDakota(comm, output_manager, transientAdjoint_print_parameters);
      }
      else
      {
        Report::UserWarning0()
          << "Sensitivity output cannot be written in requested format, using standard format";

        transientAdjoint_print_parameters.defaultExtension_ = ".TRADJ.prn";
        outputter = new Outputter::SensitivityPrn(comm, output_manager, transientAdjoint_print_parameters);
      }

      output_manager.addOutputter(PrintType::TRANADJOINT, outputter);
    }
  }
}

} // namespace Outputter
} // namespace IO
} // namespace Xyce
