//-----------------------------------------------------------------------------
// Copyright Notice
//
//   Copyright 2002 Sandia Corporation. Under the terms
//   of Contract DE-AC04-94AL85000 with Sandia Corporation, the U.S.
//   Government retains certain rights in this software.
//
//    Xyce(TM) Parallel Electrical Simulator
//    Copyright (C) 2002-2017 Sandia Corporation
//
//    This program is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program.  If not, see <http://www.gnu.org/licenses/>.
//-----------------------------------------------------------------------------

//-----------------------------------------------------------------------------
// Purpose       : Transient analysis functions.
// Special Notes :
// Creator       : Richard Schiek, SNL, Electrical and Microsystem Modeling
// Creation Date : 03/10/2009
//
//-----------------------------------------------------------------------------

#include <Xyce_config.h>

#include <N_IO_MeasureRelativeError.h>
#include <N_IO_OutputPrn.h>
#include <N_IO_OutputCsd.h>
#include <N_ERH_ErrorMgr.h>
#include <N_UTL_ExtendedString.h>
#include <N_DEV_Interpolators.h>
#include <Epetra_SerialDenseVector.h>

namespace Xyce {
namespace IO {
namespace Measure {

//-----------------------------------------------------------------------------
// Function      : RelativeError::RelativeError()
// Purpose       :
// Special Notes : 
// Scope         : public
// Creator       : Rich Schiek, Electrical and Microsystems Modeling
// Creation Date : 3/10/2009
//-----------------------------------------------------------------------------
RelativeError::RelativeError( const Util::OptionBlock & measureBlock ):
  Base(measureBlock)
{
  // indicate that this measure type is supported and should be processed in simulation
  typeSupported_ = true;

  // updateTran() and update AC() are likely to segfault if the .MEASURE line was incomplete
  checkMeasureLine();

  // Explicitly check that both column indexes are >= 0, since the default values for 
  // independentVarColumn_ and dependentVarColumn_ were set to -1 in the base object
  // constructor.  This handles incorrect values from the user, and also cases where 
  // either the INDEPVARCOL or DEPVARCOL qualifiers were not specified on the .MEASURE 
  // instance line.  Also error-out, if their values are the same. The case of either 
  // column index referring to a non-existent column in the data file is handled below, 
  // after the dataFile is read in.
  if ( (independentVarColumn_ < 0) || (dependentVarColumn_ < 0) )
  {
    Report::UserError0() << "In measure " << name_ << ", missing or negative value for INDEPVARCOL or DEPVARCVOL";
    return;
  }
  else if ( independentVarColumn_ == dependentVarColumn_ )
  {
    Report::UserError0() << "In measure " << name_ << ", identical values for INDEPVARCOL and DEPVARCVOL";
    return;
  }

  // Use the same functions for reading the .PRN, .CSD and .CSV files as re-measure.
  // The function logic is to first test for a valid extension. If the extension is invalid 
  // then the function errors out.  If the extension is valid, then make the correct object 
  // type and try to open the file.
  OutputFileBase * comparisonFile;
  ExtendedString fileExt(dataFileName_.substr(dataFileName_.length()-4));

  // The Xyce format for a .CSV file is basically the same as a .PRN file made with the
  // -delim COMMA command line option.  The only difference is that the .PRN file has something
  // like "End of Xyce(TM) Simulation" as its last line.  So, reading-in of .PRN and .CSV files 
  // currently use the same function.
  if ( (fileExt.toUpper() == ".PRN")  || (fileExt.toUpper() == ".CSV") )
  {
    comparisonFile = new OutputPrn();
  }
  else if (fileExt.toUpper() == ".CSD")
  {
    comparisonFile = new OutputCsd();
  }
  else
  {
    // unsupported file format.  Report error and exit.
    Report::UserError0() << "For measure " + name_ + ": ERROR measure only supports .PRN, .CSV or .CSD formats";
    return;
  }

  if (!comparisonFile->openFileForRead(dataFileName_))
  {
    // open failed.  Report error and exit.
    Report::UserError0() << "Could not open file in ERROR measure " + name_ + ". Failed filename was = " + dataFileName_ ;
    return;
  }

    // load data-names
  std::vector<std::string> fileVarNames;
  if (!(comparisonFile->getOutputVarNames(fileVarNames)))
  {
    // reading var names failed.  report error and exit remeasure routine.
    Report::UserError0() << "Problem reading variable names in file for measure " + name_ ;
    return;
  }

  // used for error measure generation below, during reading-in of comparison file
  double minIndepVarValues = 0.0;
  int idx=0;

  // This code section run though the lines in the comparison file, and makes the 
  // the vectors for the independent (e.g., time or frequency) and dependent values.
  int reading=1; 
  std::vector<double> varValuesVec;

  while (reading==1)
  {
    reading = comparisonFile->getOutputNextVarValuesSerial(&varValuesVec);

    if( reading == 1 )
    {
      // Throw an error if either of the requested columns do not exist in the data file.
      // This will only report once such error, for each ERROR measure instance line.
      int maxVarColumn = ( independentVarColumn_ > dependentVarColumn_ ) ? independentVarColumn_ : dependentVarColumn_;
      if( maxVarColumn >=  varValuesVec.size() )
      {
        Report::UserError0() << "In measure " << name_  << ", using data from file " << dataFileName_ 
                             << ". Requested column for independent variable " 
                             << independentVarColumn_ << " or dependent variable " 
                             << dependentVarColumn_ << " does not exist in the data file for entry " 
                             << idx;

        // exit after the first line with this error.
        return;
      }  
      else
      {
        indepVarValues_.push_back(varValuesVec[independentVarColumn_]); 
        dataValues_.push_back(varValuesVec[dependentVarColumn_]);
        if ( indepVarValues_[idx] < minIndepVarValues )
        {
          Report::UserError0() << "In measure " << name_ << ", using data from file " << dataFileName_ 
                             << ". Independent variables are not sorted in monotonically increasing order at entry " 
                             << idx;
        }
        minIndepVarValues = indepVarValues_[idx];
        idx++;
      }
      varValuesVec.clear();
    }
  }

  // results from the simulation to compare to a column in dataValues; 
  simulationDataVals_.reserve( dataValues_.size() );
  timeFreqVals_.reserve( dataValues_.size() );

  // clean up file object, which is local to this constructor
  comparisonFile->closeFileForRead();
  delete comparisonFile;
  comparisonFile = 0;
}

//-----------------------------------------------------------------------------
// Function      : RelativeError::prepareOutputVariables()
// Purpose       : Validates that the number of output variables is legal for this
//                 measure type, and then makes the vector for those variables.
// Special Notes :
// Scope         : public
// Creator       : Rich Schiek, Electrical and Microsystems Modeling
// Creation Date : 11/15/2013
//-----------------------------------------------------------------------------
void RelativeError::prepareOutputVariables()
{
  // this measurement should have only one dependent variable.
  // Error for now if it doesn't
  numOutVars_ = outputVars_.size();
  
  if ( numOutVars_ > 1 )
  {
    std::string msg = "Too many dependent variables for relative error measure, \"" + name_ + "\" Exiting.";
    Report::UserFatal() << msg;
  }

  outVarValues_.resize( numOutVars_, 0.0 );
}

//-----------------------------------------------------------------------------
// Function      : RelativeError::reset()
// Purpose       : Called when restarting a measure function.  Resets any state
// Special Notes :
// Scope         : public
// Creator       : Rich Schiek, Electrical and Microsystems Modeling
// Creation Date : 8/28/2014
//-----------------------------------------------------------------------------
void RelativeError::reset() 
{
  resetBase();

  // need to reset both vectors used in updateTran() and updateAC()
  simulationDataVals_.clear();
  timeFreqVals_.clear();
}


//-----------------------------------------------------------------------------
// Function      : RelativeError::updateTran()
// Purpose       :
// Special Notes :
// Scope         : public
// Creator       : Rich Schiek, Electrical and Microsystems Modeling
// Creation Date : 3/10/2009
//-----------------------------------------------------------------------------
void RelativeError::updateTran(Parallel::Machine comm, const double circuitTime, const Linear::Vector *solnVec, const Linear::Vector *stateVec, const Linear::Vector *storeVec, const Linear::Vector *lead_current_vector, const Linear::Vector *junction_voltage_vector, const Linear::Vector *lead_current_dqdt_vector)
{  
  if( !calculationDone_ )
  {
    timeFreqVals_.push_back( circuitTime );

    for( int i=0; i< numOutVars_; i++ )
    {
      outVarValues_[i] = getOutputValue(comm, outputVars_[i], solnVec, stateVec, storeVec, 0, lead_current_vector, junction_voltage_vector, lead_current_dqdt_vector );
      simulationDataVals_.push_back( outVarValues_[i] );
    }
    initialized_ = true;
  }
}


//-----------------------------------------------------------------------------
// Function      : RelativeError::updateDC()
// Purpose       :
// Special Notes :
// Scope         : public
// Creator       : Rich Schiek, Electrical and Microsystems Modeling
// Creation Date : 3/10/2009
//-----------------------------------------------------------------------------
void RelativeError::updateDC(Parallel::Machine comm, const std::vector<Analysis::SweepParam> & dcParamsVec, const Linear::Vector *solnVec, const Linear::Vector *stateVec, const Linear::Vector *storeVec, const Linear::Vector *lead_current_vector, const Linear::Vector *junction_voltage_vector, const Linear::Vector *lead_current_dqdt_vector)
{
}

//-----------------------------------------------------------------------------
// Function      : RelativeError::updateAC()
// Purpose       :
// Special Notes :
// Scope         : public
// Creator       : Rich Schiek, Electrical and Microsystems Modeling
// Creation Date : 3/10/2014
//-----------------------------------------------------------------------------
void RelativeError::updateAC(Parallel::Machine comm, const double frequency, const Linear::Vector * solnVec, const Linear::Vector *imaginaryVec)
{
  if( !calculationDone_ )
  {
    timeFreqVals_.push_back( frequency );

    // update our outVarValues_ vector
    for( int i=0; i< numOutVars_; i++ )
    {
      outVarValues_[i] = getOutputValue(comm, outputVars_[i], solnVec, 0, 0, imaginaryVec, 0, 0, 0 );
      simulationDataVals_.push_back( outVarValues_[i] );
    }
    initialized_ = true;
  }
}

//-----------------------------------------------------------------------------
// Function      : RelativeError::printMeasureResult()
// Purpose       : ERROR measure needs its own printMeasureResult() function
//                 because it uses parallel comms.
// Special Notes : 
// Scope         : public
// Creator       : Pete Sholander, Electrical and Microsystems Modeling
// Creation Date : 8/03/2016
//-----------------------------------------------------------------------------
std::ostream& RelativeError::printMeasureResult(std::ostream& os, bool printVerbose)
{
  basic_ios_all_saver<std::ostream::char_type> save(os);
  os << std::scientific << std::setprecision(precision_);

  if (!printVerbose)
  {
    os << name_ << " = " << this->getMeasureResult() << std::endl;
  }
  else
  {
    if (initialized_)
    {
      os << name_ << " = " << this->getMeasureResult() << std::endl ;
    }
    else
    { 
      os << name_ << " = FAILED" << std::endl;
    }

    // also print out the COMP_FUNCTION used, and whether the default value was used
    if ( usedDefaultComparisonFunc_ && (comparisonFunctionName_ == "L2NORM") )
    {
      // explicitly state if the default value was used, because invalid values
      // default to the L2NORM.  See SON Bug 839.
      os << "COMP_FUNCTION defaulted to L2NORM" << std::endl; 
    }
    else
    {
      os << "COMP_FUNCTION was " << comparisonFunctionName_ << std::endl; 
    }     
  }
  
  return os;
}

//-----------------------------------------------------------------------------
// Function      : RelativeError::printMeasureWindow
// Purpose       : prints information related to time/frequency window used.
// Special Notes :
// Scope         : public
// Creator       : Pete Sholander, Electrical and Microsystem Modeling
// Creation Date : 09/8/2016
//-----------------------------------------------------------------------------
std::ostream& RelativeError::printMeasureWindow(std::ostream& os, const double endSimTime)
{
  basic_ios_all_saver<std::ostream::char_type> save(os);
  os << std::scientific << std::setprecision(precision_);
    
  // modeStr is "Time" for TRAN mode and "Freq" for AC mode.
  std::string modeStr = (mode_ == "AC") ? "Freq" : "Time";

  // the "measurement window" for the ERROR measure is set equal to the 
  // begin and end times/frequencies of the comparison waveform
  os << "Measure Start " << modeStr << "= " << indepVarValues_.front() 
     << "\tMeasure End " << modeStr << "= " << indepVarValues_.back() << std::endl; 
  return os;
}


//-----------------------------------------------------------------------------
// Function      : RelativeError::getMeasureResult()
// Purpose       :
// Special Notes :
// Scope         : public
// Creator       : Rich Schiek, Electrical and Microsystems Modeling
// Creation Date : 3/10/2009
//-----------------------------------------------------------------------------
double RelativeError::getMeasureResult()
{
  // Only need to execute these calculations on the first pass through this 
  // function.  This function will be called twice (once for the output to 
  // the .mtX file and a second time for the output to stdout). This can improve
  // performance if the comparision file has many data points in it.
  if (!gotMeasureResult_)
  {
    // make an epetra vector to hold difference values
    Epetra_SerialDenseVector differenceVector(dataValues_.size());

    // Interpolate simulation data to input data values.
    std::vector<double> interpDataVals( dataValues_.size(), 0.0 );
    if (timeFreqVals_.size() > 0)
    {
      Xyce::Device::akima<double> interp;
      interp.init( timeFreqVals_, simulationDataVals_ );
      for (unsigned int i=0; i < dataValues_.size(); i++)
      {
        interp.eval( timeFreqVals_, simulationDataVals_, indepVarValues_[i], interpDataVals[i] );
      }
    }
    //Xyce::dout() << "Found " << numberPointsFound << " out of a possible " << dataValues.size() << std::endl;
  
    // load the difference vector
    for (unsigned int i=0 ; i<dataValues_.size(); i++)
    {
      differenceVector[i] = (interpDataVals[i]-dataValues_[i]);
    }

    // default to L2NORM for any string that is not L1NORM or INFNORM
    if( comparisonFunctionName_ == "L1NORM" )
    {
      calculationResult_ = differenceVector.Norm1();
    }
    else if( comparisonFunctionName_ == "INFNORM" )
    {
      calculationResult_ = differenceVector.NormInf();
    }
    else
    {
      calculationResult_ = differenceVector.Norm2();
    }
  }
  gotMeasureResult_=true;

  return calculationResult_;
}

} // namespace Measure
} // namespace IO
} // namespace Xyce
